"""
Test continuity of polynomial basis and its gradients along an edge on
:math:`y` line (2D) or on a face in :math:`x`-:math:`y` plane (3D) between two
elements aligned with the coordinate system, stack one on top of the other. The
evaluation occurs in several points shifted by a very small amount from the
boundary between the elements into the top and the bottom element.

For H1 space, the basis should be continuous. The components of its gradient
parallel to the edge/face should be continuous as well, while the perpendicular
component should have the same absolute value, but different sign in the top
and the bottom element.

All connectivity permutations of the two elements are tested.

The serendipity basis implementation is a pure python proof-of-concept. Its
order in continuity tests is limited to 2 on 3_8 elements to decrease the tests
run time.
"""
from itertools import product
import numpy as nm
import pytest

from sfepy.base.base import assert_
import sfepy.base.testing as tst

rsels = {
    '2_3' : 'vertices in (y > -0.1) & (y < 0.1)',
    '2_4' : 'vertices in (y > 0.9) & (y < 1.1)',
    '3_4' : 'vertices in (z > -0.1) & (z < 0.1)',
    '3_8' : 'vertices in (z > 0.9) & (z < 1.1)',
}

eps = 1e-5

shifts = {
    '2_3' : nm.array([[0.0, 0.0], [0.0, eps]], dtype=nm.float64),
    '2_4' : nm.array([[0.0, 1.0], [0.0, eps]], dtype=nm.float64),
    '3_4' : nm.array([[0.0, 0.0, 0.0], [0.0, 0.0, eps]], dtype=nm.float64),
    '3_8' : nm.array([[0.0, 0.0, 1.0], [0.0, 0.0, eps]], dtype=nm.float64),
}

def _permute_quad_face(mesh0):
    from sfepy.discrete.fem import Mesh

    coors0, ngroups0, conns0, mat_ids0, desc0 = mesh0._get_io_data()
    im = nm.arange(4, 8)
    cperms = [[0, 1, 2, 3],
              [0, 2, 1, 3],
              [0, 1, 3, 2]]
    meshes = []
    for ip, cperm in enumerate(cperms):
        tr = nm.arange(12)
        tr[im] = tr[im[cperm]]

        coors = coors0.copy()
        coors[im] = coors0[im[cperm]]
        conn = conns0[0].copy()
        conn = tr[conn]

        mesh = Mesh.from_data(mesh0.name + str(ip), coors, ngroups0,
                              [conn], mat_ids0, desc0)
        meshes.append(mesh)

    return meshes

def _get_possible_oris(geom):
    import sfepy.discrete.fem.facets as facets

    if geom in ('2_3', '2_4'):
        oris = facets.ori_line_to_iter.keys()

    elif geom == '3_4':
        oris = facets.ori_triangle_to_iter.keys()

    elif geom == '3_8':
        oris = facets._quad_ori_groups.keys() # Not ori_square_to_iter!

    else:
        oris = []

    return set(oris)

def _gen_common_data(orders, gels):
    import sfepy
    from sfepy.base.base import Struct
    from sfepy.linalg import combine
    from sfepy.discrete import FieldVariable, Integral
    from sfepy.discrete.fem import Mesh, FEDomain, Field
    from sfepy.discrete.common.global_interp import get_ref_coors

    bases = ([ii for ii in combine([['2_4', '3_8'],
                                    ['lagrange', 'serendipity', 'bernstein',
                                     'lobatto', 'sem']])]
             + [ii for ii in combine([['2_3', '3_4'],
                                      ['lagrange', 'bernstein']])])
    for geom, poly_space_base in bases:
        order = orders[geom]
        if (geom == '3_8') and (poly_space_base == 'serendipity'):
            order = 2

        tst.report('geometry: %s, base: %s, order: %d'
                   % (geom, poly_space_base, order))

        integral = Integral('i', order=order)

        aux = '' if geom in ['2_4', '3_8'] else 'z'
        mesh0 = Mesh.from_file('meshes/elements/%s_2%s.mesh' % (geom, aux),
                               prefix_dir=sfepy.data_dir)
        if (geom == '3_8'):
            meshes = _permute_quad_face(mesh0)

        else:
            meshes = [mesh0]

        gel = gels[geom]

        perms = gel.get_conn_permutations()

        qps, qp_weights = integral.get_qp(gel.surface_facet.name)
        zz = nm.zeros_like(qps[:, :1])
        qps = nm.hstack(([qps] + [zz]))

        shift = shifts[geom]
        rcoors = nm.ascontiguousarray(qps
                                      + shift[:1, :] - shift[1:, :])
        ccoors = nm.ascontiguousarray(qps
                                      + shift[:1, :] + shift[1:, :])

        all_oris = _get_possible_oris(geom)
        oris = set()

        for (ir, pr), (ic, pc), (im, mesh0) in product(
                enumerate(perms), enumerate(perms), enumerate(meshes),
        ):
            tst.report('im: %d, ir: %d, ic: %d' % (im, ir, ic))
            tst.report('pr: %s, pc: %s' % (pr, pc))

            mesh = mesh0.copy()
            conn = mesh.cmesh.get_conn(mesh0.cmesh.tdim, 0).indices
            conn = conn.reshape((mesh0.n_el, -1))
            conn[0, :] = conn[0, pr]
            conn[1, :] = conn[1, pc]

            conn2 = mesh.get_conn(gel.name)
            assert_((conn == conn2).all())

            cache = Struct(mesh=mesh)

            domain = FEDomain('domain', mesh)
            omega = domain.create_region('Omega', 'all')
            region = domain.create_region('Facet', rsels[geom], 'facet')
            field = Field.from_args('f', nm.float64, shape=1,
                                    region=omega, approx_order=order,
                                    poly_space_base=poly_space_base)
            fis = region.get_facet_indices()
            conn = mesh.cmesh.get_conn_as_graph(region.dim,
                                                region.dim - 1)
            _oris = mesh.cmesh.facet_oris[conn.indptr[fis[:, 0]]
                                          + fis[:, 1]]
            oris |= set(_oris)
            if oris == all_oris:
                break

            var = FieldVariable('u', 'unknown', field)
            tst.report('# dofs: %d' % var.n_dof)

            vec = nm.empty(var.n_dof, dtype=var.dtype)

            ps = field.poly_space

            dofs = field.get_dofs_in_region(region, merge=False)
            edofs, fdofs = nm.unique(dofs[1]), nm.unique(dofs[2])

            rrc, rcells, rstatus = get_ref_coors(field, rcoors,
                                                 cache=cache)
            crc, ccells, cstatus = get_ref_coors(field, ccoors,
                                                 cache=cache)
            assert_((rstatus == 0).all() and (cstatus == 0).all())

            yield (geom, poly_space_base, qp_weights, mesh, im, ir, ic,
                   field, ps, rrc, rcells[0], crc, ccells[0],
                   vec, edofs, fdofs)

@pytest.fixture(scope='module')
def gels():
    from sfepy.discrete.fem.geometry_element import GeometryElement

    gels = {}
    for key in ['2_3', '2_4', '3_4', '3_8']:
        gel = GeometryElement(key)
        gel.create_surface_facet()
        gels[key] = gel

    return gels

def test_partition_of_unity(gels):
    from sfepy.linalg import combine
    from sfepy.discrete import Integral, PolySpace

    ok = True
    orders = {'2_3' : 5, '2_4' : 5, '3_4' : 5, '3_8' : 5}
    bases = (
        [ii for ii in combine([['2_4', '3_8'],
                               ['lagrange', 'serendipity', 'bernstein', 'sem']]
        )]
        + [ii for ii in combine([['2_3', '3_4'],
                                 ['lagrange', 'bernstein']])]
    )

    for geom, poly_space_base in bases:
        max_order = orders[geom]
        for order in range(max_order + 1):
            if (poly_space_base == 'serendipity') and not (0 < order < 4):
                continue
            if (poly_space_base == 'sem') and not (0 < order):
                continue
            tst.report('geometry: %s, base: %s, order: %d'
                       % (geom, poly_space_base, order))

            integral = Integral('i', order=2 * order)
            coors, _ = integral.get_qp(geom)

            ps = PolySpace.any_from_args('ps', gels[geom], order,
                                         base=poly_space_base)
            vals = ps.eval_base(coors)
            _ok = nm.allclose(vals.sum(axis=-1), 1, atol=1e-14, rtol=0.0)
            tst.report('partition of unity:', _ok)
            ok = ok and _ok

    assert_(ok)

@pytest.mark.slow
def test_continuity(gels):
    ok = True
    orders = {'2_3' : 3, '2_4' : 3, '3_4' : 4, '3_8' : 3}

    bads = []
    bad_families = set()
    for (geom, poly_space_base, qp_weights, mesh, im, ir, ic,
         field, ps, rrc, rcell, crc, ccell, vec,
         edofs, fdofs) in _gen_common_data(orders, gels):

        if poly_space_base in ('lagrange', 'serendipity', 'bernstein', 'sem'):
            rbf = ps.eval_base(rrc)
            cbf = ps.eval_base(crc)

        else:
            rbf = ps.eval_base(rrc, ori=field.ori[:1])
            cbf = ps.eval_base(crc, ori=field.ori[1:])

        dofs = nm.r_[edofs, fdofs]

        res = nm.zeros((2, dofs.shape[0]), dtype=nm.int32)
        res[0, :] = dofs
        for ii, ip in enumerate(dofs):
            vec.fill(0.0)
            vec[ip] = 1.0

            evec = vec[field.econn]

            rvals = nm.dot(rbf, evec[rcell])
            cvals = nm.dot(cbf, evec[ccell])

            _ok = nm.allclose(rvals, cvals, atol=1e-14, rtol=0.0)
            res[1, ii] = _ok
            if not _ok:
                bads.append([geom, poly_space_base, im, ir, ic, ip])
                bad_families.add((geom, poly_space_base))

            ok = ok and _ok

        tst.report('results (dofs, status: 1 ok, 0 failure):\n%s' % res)

    if not ok:
        tst.report('continuity errors:\n', bads)
        tst.report('%d in total!' % len(bads))
        tst.report('continuity errors occurred in these spaces:\n',
                   bad_families)

    assert_(ok)

@pytest.mark.slow
def test_gradients(gels):
    from sfepy.discrete.fem.mappings import FEMapping

    ok = True
    orders = {'2_3' : 3, '2_4' : 3, '3_4' : 4, '3_8' : 3}

    bads = []
    bad_families = set()
    for (geom, poly_space_base, qp_weights, mesh, im, ir, ic,
         field, ps, rrc, rcell, crc, ccell, vec,
         edofs, fdofs) in _gen_common_data(orders, gels):
        gel = gels[geom]
        conn = mesh.get_conn(gel.name)

        geo_ps = field.gel.poly_space
        rmapping = FEMapping(mesh.coors, conn[rcell:rcell+1],
                             poly_space=geo_ps)
        rori = field.ori[:1] if field.ori is not None else None
        rvg = rmapping.get_mapping(rrc, qp_weights, poly_space=ps, ori=rori)
        rbfg = rvg.bfg

        cmapping = FEMapping(mesh.coors, conn[ccell:ccell+1],
                             poly_space=geo_ps)
        cori = field.ori[1:] if field.ori is not None else None
        cvg = cmapping.get_mapping(crc, qp_weights, poly_space=ps, ori=cori)
        cbfg = cvg.bfg

        dofs = nm.r_[edofs, fdofs]

        res = nm.zeros((2, dofs.shape[0]), dtype=nm.int32)
        res[0, :] = dofs
        for ii, ip in enumerate(dofs):
            vec.fill(0.0)
            vec[ip] = 1.0

            evec = vec[field.econn]

            rvals = nm.dot(rbfg, evec[rcell])[0]
            cvals = nm.dot(cbfg, evec[ccell])[0]

            okx = nm.allclose(rvals[:, 0], cvals[:, 0],
                              atol=1e-12, rtol=0.0)
            if gel.dim == 2:
                oky = nm.allclose(rvals[:, 1], -cvals[:, 1],
                                  atol=1e-12, rtol=0.0)
                _ok = okx and oky

            else:
                oky = nm.allclose(rvals[:, 1], cvals[:, 1],
                                  atol=1e-12, rtol=0.0)
                okz = nm.allclose(rvals[:, 2], -cvals[:, 2],
                                  atol=1e-12, rtol=0.0)
                _ok = okx and oky and okz

            res[1, ii] = _ok
            if not _ok:
                bads.append([geom, poly_space_base, im, ir, ic, ip])
                bad_families.add((geom, poly_space_base))

            ok = ok and _ok

        tst.report('results (dofs, status: 1 ok, 0 failure):\n%s' % res)

    if not ok:
        tst.report('gradient continuity errors:\n', bads)
        tst.report('%d in total!' % len(bads))
        tst.report('gradient continuity errors occurred in these'
                   ' spaces:\n', bad_families)

    assert_(ok)

def test_hessians(gels):
    """
    Test the second partial derivatives of basis functions using finite
    differences.
    """
    from sfepy.linalg import combine
    from sfepy.discrete import Integral, PolySpace

    ok = True
    orders = {'2_3' : 3, '2_4' : 3, '3_4' : 4, '3_8' : 3}
    bases = ([ii for ii in combine([['2_3', '2_4', '3_4', '3_8'],
                                    ['lagrange']])])

    for geom, poly_space_base in bases:
        tst.report('geometry: %s, base: %s' % (geom, poly_space_base))
        order = orders[geom]

        integral = Integral('i', order=order)
        coors, _ = integral.get_qp(geom)

        ps = PolySpace.any_from_args('ps', gels[geom], order,
                                     base=poly_space_base)

        dim = coors.shape[1]
        h1 = nm.zeros((coors.shape[0], dim, dim, ps.n_nod), nm.float64)
        eps = 1e-8
        for ir in range(dim):
            cc = coors.copy()

            cc[:, ir] -= eps
            aux0 = ps.eval_base(cc, diff=1)

            cc[:, ir] += 2 * eps
            aux1 = ps.eval_base(cc, diff=1)

            h1[:, :, ir, :] = 0.5 * (aux1 - aux0) / eps

        h2 = ps.eval_base(coors, diff=2)

        _ok = nm.allclose(h1, h2, rtol=0, atol=50*eps)
        tst.report('hessians: error: %.2e ok: %s'
                   % (nm.abs(h1 - h2).max(), _ok))
        ok = ok and _ok

    assert_(ok)



import numpy as nm
import numpy.testing as nmts

from sfepy.base.base import Struct
from sfepy.discrete.fem import FEDomain
from sfepy.mesh.mesh_generators import gen_block_mesh
from sfepy.discrete.dg.poly_spaces import get_n_el_nod
from sfepy.discrete.dg.fields import DGField


def prepare_dgfield(approx_order, mesh):

    domain = FEDomain("test_domain", mesh)
    omega = domain.create_region('Omega', 'all')
    regions = {}
    if mesh.dim > 1:
        left = domain.create_region('left',
                                    'vertices in x == 0',
                                    'edge')

        right = domain.create_region('right',
                                     'vertices in x == 1',
                                     'edge')
        bottom = domain.create_region('bottom',
                                      'vertices in y == 0',
                                      'edge')
        top = domain.create_region('top',
                                   'vertices in y == 1',
                                   'edge')
        regions.update({"top": top, "bottom": bottom})
    else:
        left = domain.create_region('left',
                                    'vertices in x == 0',
                                    'vertex')
        right = domain.create_region('right',
                                     'vertices in x == 1',
                                     'vertex')

    regions.update({"left": left, "right": right, "omega" : omega})

    field = DGField('dgfu', nm.float64, 'scalar', omega,
                    approx_order=approx_order)

    return field, regions


def prepare_dgfield_1D(approx_order):
    mesh = gen_block_mesh((1,), (4,), (.5,))
    return prepare_dgfield(approx_order, mesh), mesh


def prepare_field_2D(approx_order):
    mesh = gen_block_mesh((1, 1), (4, 4), (.5, .5))
    return prepare_dgfield(approx_order, mesh), mesh


class TestDGField:
    def test_get_facet_idx1D(self):
        mesh = gen_block_mesh((1,), (4,), (.5,))
        field, regions = prepare_dgfield(1, mesh)
        assert nm.all(field.get_bc_facet_idx(regions["left"]) == nm.array([[0, 0]]))
        assert nm.all(field.get_bc_facet_idx(regions["right"]) == nm.array([[2, 1]]))


    def test_get_facet_idx2D(self):
        mesh = gen_block_mesh((1, 1), (4, 4), (.5, .5))
        field, regions = prepare_dgfield(1, mesh)
        assert nm.all(field.get_bc_facet_idx(regions["left"]) == nm.array([[0, 3], [1, 3], [2, 3]]))
        assert nm.all(field.get_bc_facet_idx(regions["top"]) == nm.array([[2, 2], [5, 2], [8, 2]]))


    def test_create_output2D(self):
        mesh = gen_block_mesh((1, 1), (4, 4), (.5, .5))
        approx_order = 2
        n_cell_nod = 9
        field, regions = prepare_dgfield(approx_order, mesh)
        dofs = nm.zeros((n_cell_nod * 9, 1))
        output = field.create_output(dofs, "u")
        assert output["u_cell_nodes"].mode == "cell_nodes"
        assert nm.allclose(output["u_cell_nodes"].data, nm.zeros((9, n_cell_nod)))
        assert output["u_cell_nodes"].scheme is not None


    def test_create_output1D(self):
        mesh = gen_block_mesh((1,), (4,), (.5,))
        approx_order = 2
        n_cell_nod = approx_order + 1
        field, regions = prepare_dgfield(approx_order, mesh)
        dofs = nm.zeros((n_cell_nod * 3, 1))
        output = field.create_output(dofs, "u")
        for i in range(n_cell_nod):
            assert output["u_modal{}".format(i)].mode == "cell"
            assert nm.allclose(output["u_modal{}".format(i)].data, nm.zeros((3, 1)))


    def test_get_bc_facet_values_1D(self):
        (field, regions), mesh = prepare_dgfield_1D(3)
        fun = 42
        coor, val = field.get_bc_facet_values(fun, regions["left"], ret_coors=True)
        nmts.assert_equal(nm.array(0).reshape((1, 1, 1)), coor)
        nmts.assert_array_equal(val, nm.array(fun).reshape((1, 1)))

    def test_get_bc_facet_values_2D_const(self):
        (field, regions), mesh = prepare_field_2D(2)
        fun = 42
        coor, val = field.get_bc_facet_values(fun, regions["left"], ret_coors=True)
        qps = nm.array([
            [0, 1/2 * (-nm.sqrt(3/5) + 1)],
            [0, 1/2],
            [0, 1/2 * (nm.sqrt(3/5) + 1)]])
        physical_qps = nm.stack([1/3 * qps,
                                 1/3 * qps + 1/3,
                                 1/3 * qps + 2/3])
        physical_qps[:, :, 0] = 0
        nmts.assert_allclose(coor, physical_qps)
        nmts.assert_allclose(val,  nm.ones((3, 3)) * fun)

    def test_get_bc_facet_values_2D(self):
        (field, regions), mesh = prepare_field_2D(2)
        fun = lambda x: nm.sum(x ** 2, axis=-1)
        coor, val = field.get_bc_facet_values(fun, regions["left"], ret_coors=True)
        qps = nm.array([
            [0, 1 / 2 * (-nm.sqrt(3 / 5) + 1)],
            [0, 1 / 2],
            [0, 1 / 2 * (nm.sqrt(3 / 5) + 1)]])
        physical_qps = nm.stack([1 / 3 * qps,
                                 1 / 3 * qps + 1 / 3,
                                 1 / 3 * qps + 2 / 3])
        physical_qps[:, :, 0] = 0
        nmts.assert_allclose(coor, physical_qps)
        nmts.assert_allclose(val, fun(physical_qps))

    def test_set_dofs_1D(self):
        (field, regions), mesh = prepare_dgfield_1D(2)
        fun = lambda x: nm.sum(x ** 2, axis=-1)[..., None]
        nods, vals = field.set_dofs(fun, regions["omega"])
        rnods = nm.arange(3*3)
        qps = nm.array([
            [0, 1 / 2 * (-nm.sqrt(3 / 5) + 1)],
            [0, 1 / 2],
            [0, 1 / 2 * (nm.sqrt(3 / 5) + 1)]])
        physical_qps = nm.stack([1 / 3 * qps,
                                 1 / 3 * qps + 1 / 3,
                                 1 / 3 * qps + 2 / 3])

        physical_qps = physical_qps[:, :, 1]
        nmts.assert_equal(nods, rnods)
        assert vals.shape == physical_qps.shape



    def test_set_dofs_2D(self):
        (field, regions), mesh = prepare_field_2D(3)

        fun = lambda x: nm.sum(x ** 2, axis=-1)[..., None]

        nods, vals = field.set_dofs(fun, regions["omega"])
        n_el_nod = get_n_el_nod(3, 2, extended=True)
        epxected_nods = nm.arange(n_el_nod * 9)
        rvals_shape = (9, n_el_nod)
        nmts.assert_equal(nods, epxected_nods)
        assert vals.shape == rvals_shape


    def test_get_facet_neighbor_idx_1d(self):
        (field, regions), mesh = prepare_dgfield_1D(3)
        eq_map = Struct()
        eq_map.n_epbc = 0
        eq_map.dg_epbc = []
        eq_map.n_dg_epbc = 0
        nbr_idx = field.get_facet_neighbor_idx(regions["omega"], eq_map)

        rnbr_idx = [[[-1, -1], [1, 0]],  # 0
                    [[ 0,  1], [2, 0]],  # 1
                    [[ 1, 1],  [-1, -1]],  # 2
                    ]
        rnbr_idx = nm.array(rnbr_idx, dtype=nm.int32)
        nmts.assert_equal(rnbr_idx, nbr_idx)


        # periodic BCs
        eq_map.dg_epbc = [(nm.array([[0, 0]], dtype=nm.int32),
                           nm.array([[2, 1]], dtype=nm.int32))]
        field.clear_facet_neighbour_idx_cache()
        nbr_idx = field.get_facet_neighbor_idx(regions["omega"], eq_map)

        rnbr_idx = [[[2, 1], [1, 0]],  # 0
                    [[0, 1], [2, 0]],  # 1
                    [[1, 1], [0, 0]],  # 2
                    ]
        rnbr_idx = nm.array(rnbr_idx, dtype=nm.int32)
        nmts.assert_equal(rnbr_idx, nbr_idx)


    def test_get_facet_neighbor_idx_2d(self):
        (field, regions), mesh = prepare_field_2D(3)
        eq_map = Struct()
        eq_map.n_epbc = 0
        eq_map.n_dg_epbc = 0
        eq_map.dg_epbc = []
        nbr_idx = field.get_facet_neighbor_idx(regions["omega"], eq_map)

        rnbr_idx = [[[-1, -1], [3, 3], [1, 0], [-1, -1]],  # 0
                    [[ 0,  2], [4, 3], [2, 0], [-1, -1]],  # 1
                    [[ 1, 2],  [5, 3],[-1, -1],[-1, -1]],  # 2
                    [[-1, -1], [6, 3],[ 4,  0],[ 0,  1]],  # 3
                    [[ 3,  2], [7, 3],[ 5,  0], [1,  1]],  # 4
                    [[ 4, 2],  [8, 3], [-1, -1], [2, 1]],  # 5
                    [[-1, -1], [-1, -1], [7, 0], [3, 1]],  # 6
                    [ [6, 2],  [-1, -1], [8, 0], [4, 1]],  # 7
                    [[7, 2], [-1, -1], [-1, -1], [5, 1]],  # 8
                    ]
        rnbr_idx = nm.array(rnbr_idx, dtype=nm.int32)
        nmts.assert_equal(rnbr_idx, nbr_idx)

        # periodic BCs
        eq_map.dg_epbc = [(nm.array([[0, 3], [1, 3], [2, 3]], dtype=nm.int32),
                           nm.array([[6, 1], [7, 1], [8, 1]], dtype=nm.int32))]
        field.clear_facet_neighbour_idx_cache()
        nbr_idx = field.get_facet_neighbor_idx(regions["omega"], eq_map)

        rnbr_idx = [[[-1, -1], [3, 3], [1, 0], [6, 1]],  # 0
                    [[0, 2], [4, 3], [2, 0], [7, 1]],  # 1
                    [[1, 2], [5, 3], [-1, -1], [8, 1]],  # 2
                    [[-1, -1], [6, 3], [4, 0], [0, 1]],  # 3
                    [[3, 2], [7, 3], [5, 0], [1, 1]],  # 4
                    [[4, 2], [8, 3], [-1, -1], [2, 1]],  # 5
                    [[-1, -1], [0, 3], [7, 0], [3, 1]],  # 6
                    [[6, 2], [1, 3], [8, 0], [4, 1]],  # 7
                    [[7, 2], [2, 3], [-1, -1], [5, 1]],  # 8
                    ]
        rnbr_idx = nm.array(rnbr_idx, dtype=nm.int32)
        nmts.assert_equal(rnbr_idx, nbr_idx)



#!/usr/bin/env python
"""
Simple script for testing various SfePy functionality, examples not
covered by tests, and running the tests.

The script just runs the commands specified in its main() using the
`subprocess` module, captures the output and compares one or more key
words to the expected ones.

The output of failed commands is saved to 'test_install.log' file.
"""
import time
from argparse import ArgumentParser, RawDescriptionHelpFormatter
import shlex
import subprocess
import logging
import re

DEBUG_FMT = '*' * 55 + '\n%s\n' + '*' * 55

def _get_logger(filename='test_install.log'):
    """
    Convenience function to set-up output and logging.
    """
    logger = logging.getLogger('test_install.py')
    logger.setLevel(logging.DEBUG)

    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)

    file_handler = logging.FileHandler(filename)
    file_handler.setLevel(logging.DEBUG)

    logger.addHandler(console_handler)
    logger.addHandler(file_handler)

    return logger

logger = _get_logger()

def check_output(cmd):
    """
    Run the specified command and capture its outputs.

    Returns
    -------
    out : tuple
        The (stdout, stderr) output tuple.
    """
    logger.info(cmd)
    args = shlex.split(cmd)

    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out = [ii.decode() for ii in p.communicate()]

    return out

def report(out, name, line, item, value, eps=None, return_item=False,
           match_numbers=False):
    """
    Check that `item` at `line` of the output string `out` is equal
    to `value`. If not, print the output.
    """
    try:
        if match_numbers:
            status = out.split('\n')[line]

        else:
            status = out.split('\n')[line].split()

    except IndexError:
        logger.error('  not enough output from command!')
        ok = False

    else:
        try:
            if match_numbers:
                pat = '([-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?[jJ]?)'
                matches = re.findall(pat, status)
                status_item = matches[item]

            else:
                status_item = status[item]

            logger.info('  comparing: %s %s', status_item, value)

            if eps is None:
                ok = (status_item == value)

            else:
                try:
                    ok = abs(float(status_item) - float(value)) < eps

                except:
                    ok = False

        except IndexError:
            ok = False

    logger.info('  %s: %s', name, ok)

    if not ok:
        logger.debug(DEBUG_FMT, out)

    if return_item:
        return ok, status[item]

    else:
        return ok

def report2(out, name, items, return_item=False):
    """
    Check that `items` are in the output string `out`.
    If not, print the output.
    """
    ok = True
    for s in items:
        logger.info('  checking: %s', s)
        if s not in out:
            ok = False
            break

    logger.info('  %s: %s', name, ok)

    if not ok:
        logger.debug(DEBUG_FMT, out)

    if return_item:
        return ok, s

    else:
        return ok

def report_tests(out, return_item=False):
    """
    Check that all tests in the output string `out` passed.
    If not, print the output.
    """
    from pyparsing import (Word, Combine, Suppress, Optional, OneOrMore,
                           delimitedList, nums, Literal)
    from functools import partial
    integer = Word(nums).setName('integer')
    real = Combine(Word(nums) + '.' + Optional(Word(nums))).setName('real')

    equals = Suppress(OneOrMore('='))
    _stats = {}
    def add_stat(s, loc, toks, key=None):
        if key is None:
            key = toks[1]
        _stats[key] = toks[0]
        return toks

    word = ((integer + 'failed') |
            (integer + 'passed') |
            (integer + 'deselected') |
            (integer + 'warnings')).setParseAction(add_stat)

    line = (equals +
            Optional(delimitedList(word)) +
            'in' +
            (real + (Literal('s') | 'seconds'))
            .setParseAction(partial(add_stat, key='seconds')) +
            equals)

    line.searchString(out)

    keys = ['failed', 'passed', 'deselected', 'warnings', 'seconds']
    stats = {key : _stats.get(key, '0') for key in keys}

    ok = stats['failed'] == '0'

    logger.info(
        ('  {failed} failed, {passed} passed, {deselected} deselected,'
         ' {warnings} warnings in {seconds} seconds').format(**stats)
    )

    if not ok:
        logger.debug(DEBUG_FMT, out)

    if return_item:
        return ok, stats['failed']

    else:
        return ok

def main():
    parser = ArgumentParser(description=__doc__,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('--version', action='version', version='%(prog)s')
    parser.parse_args()

    fd = open('test_install.log', 'w')
    fd.close()

    eok = 0

    t0 = time.time()

    out, err = check_output('python3 sfepy/scripts/blockgen.py')
    eok += report(out, '...', -2, 1, '...done')

    out, err = check_output('python3 sfepy/scripts/cylindergen.py')
    eok += report(out, '...', -2, 1, '...done')

    out, err = check_output('python3 sfepy/scripts/convert_mesh.py meshes/3d/cylinder.vtk out.mesh')
    eok += report(out, '...', -2, 1, '...done')

    out, err = check_output('python3 sfepy/scripts/convert_mesh.py --tile 2,2 meshes/elements/2_4_2.mesh out-per.mesh')
    eok += report(out, '...', -2, 1, '...done')

    out, err = check_output('python3 sfepy/scripts/convert_mesh.py --extract-surface --print-surface=- meshes/various_formats/octahedron.node surf_octahedron.mesh')
    eok += report(out, '...', -4, 0, '1185')

    out, err = check_output("""python3 sfepy/scripts/simple.py -c "ebc_2 : {'name' : 't2', 'region' : 'Gamma_Right', 'dofs' : {'t.0' : -5.0}}" sfepy/examples/diffusion/poisson.py""")
    eok += report(out, '...', -8, 5, '2.308051e-16', eps=1e-15)

    out, err = check_output('python3 sfepy/scripts/simple.py sfepy/examples/diffusion/poisson_parametric_study.py')
    eok += report(out, '...', -8, 5, '1.606408e-14', eps=1e-13)

    out, err = check_output('python3 sfepy/scripts/simple.py sfepy/examples/linear_elasticity/its2D_3.py')
    eok += report(out, '...', -29, 5, '3.964886e-12', eps=1e-11)
    eok += report(out, '...', -9, 4, '2.58660e+01', eps=1e-5)

    out, err = check_output('python3 sfepy/scripts/simple.py sfepy/examples/linear_elasticity/linear_elastic.py --format h5')
    eok += report(out, '...', -8, 5, '4.638192e-18', eps=1e-15)

    out, err = check_output('python3 sfepy/scripts/extractor.py -d cylinder.h5')
    eok += report(out, '...', -2, 1, '...done')

    out, err = check_output('python3 sfepy/scripts/resview.py --off-screen -o cylinder.png cylinder.h5')
    eok += report(out, '...', -2, 1, 'cylinder.png')

    out, err = check_output('python3 sfepy/scripts/simple.py sfepy/examples/phononic/band_gaps.py')
    eok += report(out, '...', -9, 0, '2.08545116e+08', match_numbers=True)
    eok += report(out, '...', -8, 1, '1.16309223e+11', match_numbers=True)

    out, err = check_output('python3 sfepy/scripts/simple.py sfepy/examples/phononic/band_gaps.py --phonon-phase-velocity')
    eok += report(out, '...', -2, 0, '4189.41229592', match_numbers=True)
    eok += report(out, '...', -2, 1, '2620.55608256', match_numbers=True)

    out, err = check_output('python3 sfepy/scripts/simple.py sfepy/examples/phononic/band_gaps.py --phonon-dispersion')
    eok += report(out, '...', -6, 1, '[0,')

    out, err = check_output('python3 sfepy/scripts/simple.py sfepy/examples/phononic/band_gaps_rigid.py')
    eok += report(out, '...', -9, 0, '4.58709531e+07', match_numbers=True)
    eok += report(out, '...', -8, 1, '1.13929200e+11', match_numbers=True)

    out, err = check_output('python3 sfepy/scripts/simple.py sfepy/examples/quantum/hydrogen.py')
    eok += report(out, '...', -2, -2, '-0.01913506', eps=1e-4)

    out, err = check_output('python3 sfepy/scripts/simple.py sfepy/examples/homogenization/perfusion_micro.py')
    eok += report2(out, '...', ['computing EpA', 'computing PA_3',
                                'computing GA', 'computing EmA',
                                'computing KA'])

    out, err = check_output('python3 sfepy/examples/homogenization/rs_correctors.py -n')
    eok += report(out, '...', -2, -1, '1.644e-01', match_numbers=True)

    out, err = check_output('python3 sfepy/examples/large_deformation/compare_elastic_materials.py -n')
    eok += report(out, '...', -8, 5, '1.068759e-14', eps=1e-13)

    out, err = check_output('python3 sfepy/examples/linear_elasticity/linear_elastic_interactive.py')
    eok += report(out, '...', -18, 0, '1.62128841139e-14', eps=1e-13)

    out, err = check_output('python3 sfepy/examples/linear_elasticity/elastodynamic_identification.py --opt-conf=xtol=0.5')
    eok += report(out, '...', -5, 0, '2', match_numbers=True)

    out, err = check_output('python3 sfepy/examples/linear_elasticity/modal_analysis.py')
    eok += report(out, '...', -12, 5, '12142.11470773', eps=1e-13)

    out, err = check_output('python3 sfepy/examples/multi_physics/thermal_electric.py')
    eok += report(out, '...', -9, 5, '2.612933e-14', eps=1e-13)

    out, err = check_output('python3 sfepy/examples/diffusion/laplace_refine_interactive.py output')
    eok += report(out, '...', -3, 5, '2.675866e-15', eps=1e-13)

    out, err = check_output('python3 sfepy/examples/diffusion/laplace_iga_interactive.py -o output')
    eok += report(out, '...', -3, 5, '1.028134e-13', eps=1e-12)

    out, err = check_output('python3 sfepy/examples/dg/imperative_burgers_1D.py -o output')
    eok += report(out, '...', -3, 3, 'moment_1D_limiter')

    out, err = check_output('mpiexec -n 2 python3 sfepy/examples/diffusion/poisson_parallel_interactive.py output-parallel -2 --silent -ksp_monitor')
    eok += report(out, '...', -2, 4, '8.021313824020e-07', eps=1e-6)

    out, err = check_output('mpiexec -n 2 python3 sfepy/examples/multi_physics/biot_parallel_interactive.py output-parallel -2 --silent -ksp_monitor')
    eok += report(out, '...', -2, 4, '3.787214380277e-09', eps=1e-7)

    t1 = time.time()

    out, err = check_output("python3 -c \"import sfepy; sfepy.test('-v', '--disable-warnings')\"")
    tok, failed = report_tests(out, return_item=True)
    tok = {True : 'ok', False : 'fail'}[tok]

    t2 = time.time()

    fd = open('test_install_times.log', 'a+')
    fd.write('%s: examples: %.2f [s] (%d), tests: %.2f [s] (%s: %s)\n'
             % (time.ctime(t0), t1 - t0, eok, t2 - t1, tok, failed))
    fd.close()

if __name__ == '__main__':
    main()



import numpy as nm

from sfepy.terms.terms_multilinear import ETermBase

class MassTerm(ETermBase):
    r"""
    Mass term with lumping and RMM support [1]_.

    The lumping parameter can be 'row_sum', 'hrz' or 'none' (default). It
    applies for :math:`\beta` > 0:

    - :math:`\beta` = 0 correponds to the consistent mass matrix :math:`M^C`;
    - 0 < :math:`\beta` < 1 corresponds to the averaged mass matrix
      :math:`M^A`.
    - :math:`\beta` = 1 corresponds to the lumped mass matrix :math:`M^L`;

    `term_mode` can be None (default), 'DPM' (diagonal projection matrix
    :math:`A`), or 'RMM' (reciprocal mass matrix :math:`C`).

    .. [1] González, J.A., Kolman, R., Cho, S.S., Felippa, C.A., Park, K.C.,
           2018. Inverse mass matrix via the method of localized Lagrange
           multipliers. International Journal for Numerical Methods in
           Engineering 113, 277–295. https://doi.org/10.1002/nme.5613

    :Definition:

    .. math::
        M^C = \int_{\cal{D}} \rho \ul{v} \cdot \ul{u} \\
        M^L = \mathrm{lumping}(M^C) \\
        M^A = (1 - \beta) M^C + \beta M^L \\
        A = \sum_e A_e \\
        C = \sum_e A_e^T (M_e^A)^{-1} A_e

    :Arguments:
        - material: :math:`\rho`
        - material: lumping
        - material: :math:`\beta`
        - virtual/parameter_1: :math:`\ul{v}`
        - state/parameter_2: :math:`\ul{u}`
    """
    name = 'de_mass'
    arg_types = ('material_rho', 'material_lumping', 'material_beta',
                 'virtual', 'state')
    arg_shapes = {'material_rho' : '1, 1', 'material_lumping' : '.: str',
                  'material_beta' : '.: 1',
                  'virtual' : ('D', 'state'), 'state' : 'D'}
    modes = ('weak', 'eval')

    def get_function(self, rho, lumping, beta, virtual, state,
                     mode=None, term_mode=None, diff_var=None, **kwargs):

        if ((term_mode is None)
            and ((lumping == 'none') or (beta == 0.0))):
            # Consistent mass matrix.
            fun = self.make_function(
                '00,i,i', rho, virtual, state, diff_var=diff_var,
            )
            return fun

        _fun = self.make_function(
            '00,i,i', rho, virtual, state, diff_var=state.name,
        )
        self.einfos[None] = self.einfos[state.name]

        if (term_mode in ('RMM', 'DPM')) and (lumping == 'none'):
            lumping = 'row_sum'

        if (term_mode == 'DPM'):
            beta = 1.0

        def fun(out, *fargs):
            # Compute M_C.
            if diff_var is None:
                sh = out.shape
                _out = nm.zeros((sh[0], sh[1], sh[2], sh[2]), dtype=out.dtype)

            else:
                _out = out
            status = _fun(_out, *fargs)

            # Consistent element mass matrices M_C.
            mc = _out if beta in (0.0, 1.0) else _out.copy()
            # Diagonals of lumped element mass matrices M_L.
            if lumping == 'row_sum':
                mld = mc.sum(-1)

            elif lumping == 'hrz':
                n_bf = mc.shape[-1] // self.region.dim
                det = fargs[2][0][0]
                jrho = det * rho[..., 0, 0]

                diag = nm.einsum('...ii->...i', mc)
                intrho = jrho.sum(axis=1) # Element masses to preserve.
                intdiag = diag[:, 0, :n_bf].sum(axis=1)
                alpha = intrho / intdiag
                mld = alpha[:, None, None] * diag

            else:
                raise ValueError('unknown lumping mode! (%s)' % lumping)

            if beta == 1.0:
                # M_A = M_L.
                eye = nm.eye(mld.shape[-1])
                _out[:] = nm.einsum('can,nm->canm', mld, eye)

            elif beta > 0.0:
                # M_A = (1 - beta) * M_C + beta * M_L.
                _out[:] = (1.0 - beta) * mc
                outd = nm.einsum('...ii->...i', _out) # View to diagonal.
                outd += beta * mld

            else:
                # beta == 0.0, so M_A = M_C in RMM term_mode -> do nothing.
                pass

            if term_mode == 'RMM':
                # out contains M_A, mld contains A_e diagonal.
                ime = nm.linalg.inv(_out)

                # ce = nm.einsum('cik,ckl,clj->cij', ae, ime, ae)
                _out[:] =  mld[..., None] * ime * mld[:, None, :]

            if diff_var is None:
                earg = self.einfos[None].eargs[2]
                step_cache = earg.arg.evaluate_cache.setdefault('dofs', {})
                cache = step_cache.setdefault(0, {})
                dofs = earg.get_dofs(cache, step_cache, state.name + '.dofs')
                dofs = dofs.reshape((sh[0], sh[2])) # Does a copy!

                # Could be faster by exploiting that mld is a vector.
                out[..., 0] = nm.einsum('caij,cj->cai', _out, dofs)

            return status

        return fun



import os, glob

from .config import Config
from .version import __version__, in_source_tree, top_dir

data_dir = os.path.realpath(top_dir)
base_dir = os.path.dirname(os.path.normpath(os.path.realpath(__file__)))

def get_paths(pattern):
    """
    Get files/paths matching the given pattern in the sfepy source tree.
    """
    if not in_source_tree:
        pattern = '../' + pattern

    files = glob.glob(os.path.normpath(os.path.join(top_dir, pattern)))
    return files

def test(*args):
    """
    Run all the package tests.

    Equivalent to running ``pytest sfepy/tests/`` in the base directory of
    SfePy. Allows an installed version of SfePy to be tested.

    To test an installed version of SfePy use

    .. code-block:: bash

       $ python -c "import sfepy; sfepy.test()"

    Parameters
    ----------
    *args : positional arguments
        Arguments passed to pytest.
    """
    import pytest  # pylint: disable=import-outside-toplevel

    args = list(args)
    if all(arg.startswith('-') for arg in args):
        # Add the default path only if no path is given explicitly.
        path = os.path.join(os.path.split(__file__)[0], 'tests')
        args = [path] + args

    return pytest.main(args=args)



r"""
An example demonstrating the usage of the truss elements in 2D.
The bridge structure is fixed on the left and supported on the right.

Running the simulation::

  sfepy-run sfepy/examples/linear_elasticity/truss_bridge.py

Viewing the results::

  sfepy-view bridge.vtk -f u:wu:p0 1:vw:p0 S:e:p1 --2d-view
"""
import numpy as nm
from sfepy.base.base import Struct
from sfepy import data_dir


def post_process(out, pb, state, extend=False):
    valS = pb.evaluate('ev_lin_truss_force.0.Truss(truss.EA, u)', mode='el_avg')
    out['S'] = Struct(name='output_data', mode='cell',
                      region_name='Truss', data=valS)

    print('### nodal displacements: ')
    print(out['u'].data)

    return out


def get_pars(ts, coors, mode=None, **kwargs):
    if mode == 'qp':
        EA = nm.array([2, 2, 2, 2, 2, 2,
                       10, 10, 10, 10, 10, 10,
                       3, 3, 3, 3, 3,
                       1, 1, 1, 1]).reshape(-1, 1, 1) * 1000

        return {'EA': EA}


filename_mesh = data_dir + "/meshes/2d/bridge.vtk"

options = {
    'post_process_hook': 'post_process',
}

regions = {
    'Truss': 'cells of group 1',
    'Left': ('vertices in (x < 0.01)', 'vertex'),
    'Right': ('vertices in (x > 59.99)', 'vertex'),
    'Bottom': ('vertex 2, 4, 6, 8, 10', 'vertex'),
}

materials = {
    'force': ({'.val': nm.array([[0, -10],
                                 [0, -10],
                                 [0, -16],
                                 [0, -10],
                                 [0, -10]])},),
    'truss': 'get_pars',
}

functions = {
    'get_pars': (get_pars,),
}

fields = {
    'displacement': ('real', 'vector', 'Truss', 1),
}

variables = {
    'u': ('unknown field', 'displacement'),
    'v': ('test field', 'displacement', 'u'),
}

ebcs = {
    'Fixed_Left': ('Left', {'u.all': 0.0}),
    'Support_Right': ('Right', {'u.1': 0.0}),
}

equations = {
    'balance_of_forces':
    """dw_lin_truss.0.Truss(truss.EA, v, u)
     = dw_point_load.0.Bottom(force.val, v)""",
}

solvers = {
    'ls': ('ls.auto_direct', {}),
    'newton': ('nls.newton', {}),
}



import gc
import numpy as nm
import pytest

from sfepy.mechanics.matcoefs import stiffness_from_youngpoisson
from sfepy import data_dir
import sfepy.base.testing as tst

filename_mesh = data_dir + '/meshes/3d/special/cube_sphere.mesh'

fields = {
    'scalar_field' : ('real', 'scalar', 'Omega', 1),
    'vector_field' : ('real', 'vector', 'Omega', 1),
}

variables = {
    'U1': ('parameter field', 'vector_field', '(set-to-None)'),
    'U2': ('parameter field', 'vector_field', '(set-to-None)'),
    'P1': ('parameter field', 'scalar_field', '(set-to-None)'),
    'P2': ('parameter field', 'scalar_field', '(set-to-None)'),
    'V': ('parameter field', 'vector_field', '(set-to-None)'),
}

regions = {
    'Omega': 'all',
    'Omega1': 'cells of group 1',
    'Omega2': 'cells of group 2',
    'Interface': ('r.Omega1 *v r.Omega2', 'facet', 'Omega1'),
}

integrals = {
    'i': 2,
}

piezo_g = nm.array([[0, 0, 0, 0, 10, 0],
                    [0, 0, 0, 0, 0, 10],
                    [-5, -5, 15, 0, 0, 0]])

materials = {
    'mat': ({
        'D': {
            'Omega1': stiffness_from_youngpoisson(3, 100, 0.3),
            'Omega2': stiffness_from_youngpoisson(3, 1, 0.49),
        },
        'K': {'Omega1': nm.eye(3), 'Omega2': nm.eye(3) * 1e-3},
        'c': {'Omega1': 1, 'Omega2': 1000},
        'g': {'Omega1': piezo_g * 1e2, 'Omega2': piezo_g},
        's': {
            'Omega1': nm.array([[1, 2, 3, 0, 0, 0]]).T,
            'Omega2': nm.array([[8, 4, 2, 0, 0, 0]]).T,
        },
    },),
}

equations = {}

test_terms = [
    ('ev_sd_lin_elastic', 'dw_lin_elastic', 'Omega', 'mat.D', 'U1', 'U2'),
    ('de_sd_lin_elastic', 'dw_lin_elastic', 'Omega', 'mat.D', 'U1', 'U2'),
    ('ev_sd_dot', 'dw_dot', 'Omega', None, 'U1', 'U2'),
    ('de_sd_dot', 'de_dot', 'Omega', 'mat.K', 'U1', 'U2'),
    ('de_sd_dot', 'de_dot', 'Omega', 'mat.c', 'P1', 'P2'),
    ('ev_sd_diffusion', 'dw_diffusion', 'Omega', 'mat.K', 'P1', 'P2'),
    ('de_sd_diffusion', 'de_diffusion', 'Omega', 'mat.K', 'P1', 'P2'),
    ('ev_sd_div', 'dw_stokes', 'Omega', None, 'U1', 'P1'),
    ('de_sd_stokes', 'de_stokes', 'Omega', None, 'U1', 'P1'),
    ('ev_sd_div_grad', 'dw_div_grad', 'Omega', 'mat.c', 'U1', 'U2'),
    ('de_sd_div_grad', 'de_div_grad', 'Omega', 'mat.c', 'U1', 'U2'),
    ('ev_sd_piezo_coupling', 'dw_piezo_coupling', 'Omega', 'mat.g', 'U1', 'P1'),
    ('de_sd_piezo_coupling', 'dw_piezo_coupling', 'Omega', 'mat.g', 'U1', 'P1'),
    ('ev_sd_convect', 'de_convect', 'Omega', None, 'U1', 'U2'),
    ('ev_sd_surface_ltr', 'dw_surface_ltr', 'Interface', 'mat.s', 'U1', None),
    ('de_sd_surface_ltr', 'de_surface_ltr', 'Interface', 'mat.s', 'U1', None),
    ('de_sd_surface_ltr', 'de_surface_ltr', 'Interface', 'mat.c', 'U1', None),
    ('de_sd_v_dot_grad_s', 'dw_v_dot_grad_s', 'Omega', 'mat.c', 'U1', 'P1'),
]


def modify_mesh(val, spbox, dv_mode, cp_pos):
    cpoints0 = spbox.get_control_points().copy()
    for pts, dir in dv_mode:
        for pt in pts:
            spbox.move_control_point(cp_pos[pt], nm.array(dir) * val)

    new_coors = spbox.evaluate()
    spbox.set_control_points(cpoints0)

    return new_coors


@pytest.fixture(scope='module')
def problem():
    import sys
    from sfepy.discrete import Problem
    from sfepy.base.conf import ProblemConf

    conf = ProblemConf.from_dict(globals(), sys.modules[__name__])

    problem = Problem.from_conf(conf, init_equations=False)
    return problem


@pytest.mark.slow
def test_sensitivity(problem):
    from sfepy.discrete import Variables
    from sfepy.mesh.splinebox import SplineBox

    tolerance = 1e-4
    ok = True

    variables = Variables.from_conf(problem.conf.variables, problem.fields)

    for var_name in variables.names:
        var = variables[var_name]
        n_dof = var.field.n_nod * var.field.shape[0]
        aux = nm.arange(n_dof, dtype=nm.float64)
        var.set_data(aux)

    mesh = problem.domain.mesh
    bbox = nm.array(mesh.get_bounding_box()).T
    spbox = SplineBox(bbox, mesh.coors)

    dvel_modes = [
        # expand inner cylinder, no volume change
        [([20, 21, 22, 23], (-1, -1, 0)),
         ([24, 25, 26, 27], (-1, 1, 0)),
         ([36, 37, 38, 39], (1, -1, 0)),
         ([40, 41, 42, 43], (1, 1, 0))],
        # volume change
        [(range(16, 32), (0.2, 0, 0)),
         (range(32, 48), (0.4, 0, 0)),
         (range(48, 52), (0.6, 0.2, 0.2)),
         (range(52, 56), (0.8, 0.2, 0.3)),
         (range(56, 60), (1.0, 0.2, 0.4)),
         (range(60, 64), (1.2, 0.2, 0.5))],
    ]

    r4 = range(4)
    cp_pos = {i*16 + j*4 + k: (i, j, k)
        for k in r4 for j in r4 for i in r4}

    # compute design velocities
    dvels = []
    for dv_mode in dvel_modes:
        dvel = 0
        for pts, dir in dv_mode:
            for pt in pts:
                dvel += spbox.evaluate_derivative(cp_pos[pt], dir)
        dvels.append(dvel)

    for tname_sa, tname, rname, mat, var1, var2 in test_terms:
        args = [] if mat is None else [mat]
        args += [var1] if var2 is None else [var1, var2]
        term = '%s.i.%s(%s)' % (tname, rname, ', '.join(args))
        term_sa = '%s.i.%s(%s)' % (tname_sa, rname, ', '.join(args + ['V']))

        val = problem.evaluate(term, var_dict=variables.as_dict())
        tst.report('%s: %s' % (tname, val))

        dt = 1e-6
        for ii, dvel in enumerate(dvels):
            val = problem.evaluate(term, var_dict=variables.as_dict())
            variables['V'].set_data(dvel)
            val_sa = problem.evaluate(term_sa, var_dict=variables.as_dict())
            tst.report('%s - mesh_velocity mode %d' % (tname_sa, ii))
            # mesh perturbation +
            new_coors = modify_mesh(dt/2., spbox, dvel_modes[ii], cp_pos)
            problem.set_mesh_coors(new_coors, update_fields=True)
            val1 = problem.evaluate(term, var_dict=variables.as_dict())

            # mesh perturbation -
            new_coors = modify_mesh(-dt/2., spbox, dvel_modes[ii], cp_pos)
            problem.set_mesh_coors(new_coors, update_fields=True)
            val2 = problem.evaluate(term, var_dict=variables.as_dict())

            val_fd = (val1 - val2) / dt
            err = nm.abs(val_sa - val_fd) / nm.linalg.norm(val_sa)
            tst.report('term:               %s' % val)
            tst.report('sensitivity term:   %s' % val_sa)
            tst.report('finite differences: %s' % val_fd)
            tst.report('relative error:     %s' % err)

            _ok = err < tolerance

            ok = ok and _ok

            gc.collect()

    assert ok



r"""
The linear elastodynamics of a piezoelectric body loaded by a given base
motion.

The generated voltage between the bottom and top surface electrodes is recorded
and plotted. The scalar potential on the top surface electrode is modeled using
a constant L^2 field. The Nitsche's method is used to weakly apply the
(unknown) potential Dirichlet boundary condition on the top surface.

Find the displacements :math:`\ul{u}(t)`, the potential :math:`p(t)` and the
constant potential on the top electrode :math:`\bar p(t)` such that:

.. math::
    \int_\Omega \rho\ \ul{v} \cdot \ul{\ddot u}
    + \int_\Omega C_{ijkl}\ \veps_{ij}(\ul{v}) \veps_{kl}(\ul{u})
    - \int_\Omega e_{kij}\ \veps_{ij}(\ul{v}) \nabla_k p
    = 0
    \;, \quad \forall \ul{v} \;,

    \int_\Omega e_{kij}\ \veps_{ij}(\ul{u}) \nabla_k q
    + \int_\Omega \kappa_{ij} \nabla_i \psi \nabla_j p
    - \int_{\Gamma_{top}} \kappa_{ij} \nabla_j p n_i q
    + \int_{\Gamma_{top}} \kappa_{ij} \nabla_j q n_i (p - \bar p)
    + \int_{\Gamma_{top}} k q (p - \bar p)
    = 0
    \;, \quad \forall q \;,

    \int_{\Gamma_{top}} \kappa_{ij} \nabla_j \dot{p} n_i + \bar p / R = 0 \;,

where :math:`C_{ijkl}` is the matrix of elastic properties under constant
electric field intensity, :math:`e_{kij}` the piezoelectric modulus,
:math:`\kappa_{ij}` the permittivity under constant deformation, :math:`k` a
penalty parameter and :math:`R` the external circuit resistance (e.g. of an
oscilloscope used to measure the voltage between the electrodes).

Usage Examples
--------------

Run with the default settings, results stored in ``output/piezo-ed/``::

  sfepy-run sfepy/examples/multi_physics/piezo_elastodynamic.py

The :func:`define()` arguments, see below, can be set using the ``-d`` option::

  sfepy-run sfepy/examples/multi_physics/piezo_elastodynamic.py -d "order=2, ct1=2.5"

View the resulting potential :math:`p` on a deformed mesh (2000x magnified)::

  sfepy-view output/piezo-ed/user_block.h5 -f p:wu:f2000:p0 1:vw:wu:f2000:p0 --color-map=inferno
"""
from functools import partial

import numpy as nm

from sfepy.base.base import output
from sfepy.discrete.fem.meshio import UserMeshIO
from sfepy.mesh.mesh_generators import gen_block_mesh
from sfepy.homogenization.utils import define_box_regions

def post_process(out, problem, state, extend=False, pcs=None):
    """
    Calculate and output strain, stress and electric field vector for the given
    displacements and potential.
    """
    from sfepy.base.base import Struct

    ev = problem.evaluate
    strain = ev('ev_cauchy_strain.i.Omega(u)', mode='el_avg', verbose=False)
    stress = ev('ev_cauchy_stress.i.Omega(m.C, u)', mode='el_avg',
                copy_materials=False, verbose=False)
    E = ev('ev_grad.i.Omega(p)', mode='el_avg', verbose=False)

    out['cauchy_strain'] = Struct(name='output_data', mode='cell',
                                  data=strain)
    out['cauchy_stress'] = Struct(name='output_data', mode='cell',
                                  data=stress)
    out['E'] = Struct(name='output_data', mode='cell', data=E)

    top = problem.domain.regions['Top']
    p_top = state['p'].get_state_in_region(top)
    # = state['pc'](), but we want to test .get_state_in_region()
    pc_top = state['pc'].get_state_in_region(top)

    output('pc:', pc_top)
    output('|p - pc|_top:', nm.linalg.norm(p_top - pc_top))
    if pcs is not None:
        pcs.append(pc_top[0, 0])

    return out

def plot_voltage(problem, state, pcs=None):
    import os.path as op
    import matplotlib.pyplot as plt

    ts = problem.get_timestepper()

    fig, ax = plt.subplots()
    ax.plot(ts.times, pcs)
    ax.set_xlabel('$t$ [s]')
    ax.set_ylabel(r'$\bar p$ [V]')

    fig.tight_layout()
    fig.savefig(op.join(problem.output_dir, 'voltage.pdf'))

def define(
        dims=(1e-2, 1e-2, 5e-3),
        shape=(5, 11, 21),
        order=1,
        amplitude=0.0000001,
        ct1=1.5,
        dt=None,
        tss_name='tsn',
        tsc_name='tscedl',
        adaptive=False,
        ls_name='lsd',
        active_only=False,
        save_times='all',
        output_dir='output/piezo-ed',
):
    """
    Parameters
    ----------
    dims: physical dimensions of the block mesh
    shape: numbers of mesh vertices along each axis
    order: the FE approximation order
    amplitude: the seismic load amplitude
    ct1: final time in min(dims) / "longitudinal wave speed" units
    dt: time step (None means automatic)
    tss_name: time stepping solver name (see "solvers" section)
    tsc_name: time step controller name (see "solvers" section)
    adaptive: use adaptive time step control
    ls_name: linear system solver name (see "solvers" section)
    save_times: number of solutions to save
    output_dir: output directory
    """
    dim = len(dims)
    assert dim == 3

    # A PZT 5-H material, Voigt notation, strain - electric displacement form.
    epsT = nm.array([[1700., 0, 0],
                     [0, 1700., 0],
                     [0, 0, 1450.0]])
    dv = 1e-12 * nm.array([[0, 0, 0, 0, 741., 0],
                           [0, 0, 0, 741, 0, 0],
                           [-274., -274., 593., 0, 0, 0]]) # C / N = m / V

    # Convert to stress - electric displacement form.
    CEv = nm.array([[1.27e+011, 8.02e+010, 8.47e+010, 0, 0, 0],
                    [8.02e+010, 1.27e+011, 8.47e+010, 0, 0, 0],
                    [8.47e+010, 8.47e+010, 1.17e+011, 0, 0, 0],
                    [0, 0, 0, 2.34e+010, 0, 0],
                    [0, 0, 0, 0, 2.30e+010, 0],
                    [0, 0, 0, 0, 0, 2.35e+010]])
    ev = dv @ CEv
    epsS = epsT - dv @ ev.T

    # SfePy: 11 22 33 12 13 23
    # Voigt: 11 22 33 23 13 12
    ii = [0, 1, 2, 5, 4, 3]
    ix, iy = nm.meshgrid(ii, ii, sparse=True)
    CE = CEv[ix, iy]
    e = ev[:, ii]

    eps0 = 8.8541878128e-12
    kappa = epsS * eps0

    # Longitudinal and shear wave propagation speeds.
    mu = CE[-1, -1]
    lam = CE[0, 0] - 2 * mu
    rho = 7800
    cl = nm.sqrt((lam + 2.0 * mu) / rho)
    cs = nm.sqrt(mu / rho)

    # Element size.
    L = nm.min(dims)
    H = L / (nm.max(shape) - 1)

    # Time-stepping parameters.
    if dt is None:
        # For implicit schemes, dt based on the Courant number C0 = dt * cl / H
        # equal to 1.
        dt = H / cl # C0 = 1

    t1 = ct1 * L / cl

    # Time history record of pc.
    pcs = []
    _post_process = partial(post_process, pcs=pcs)
    _plot_voltage = partial(plot_voltage, pcs=pcs)

    def mesh_hook(mesh, mode):
        """
        Generate the block mesh.
        """
        if mode == 'read':
            mesh = gen_block_mesh(dims, shape, 0.5 * nm.array(dims),
                                  name='user_block', verbose=False)
            return mesh

        elif mode == 'write':
            pass

    filename_mesh = UserMeshIO(mesh_hook)

    bbox = [[0] * dim, dims]
    regions = define_box_regions(dim, bbox[0], bbox[1], 1e-5)
    regions.update({
        'Omega' : 'all',
    })

    materials = {
        'inclusion' : (None, 'get_inclusion_pars')
    }

    fields = {
        'displacement' : ('real', 'vector', 'Omega', order),
        'potential' : ('real', 'scalar', 'Omega', order),
        'constant' : ('real', 'scalar', 'Top', 0, 'L2', 'constant'),
    }

    variables = {
        'u' : ('unknown field', 'displacement', 0),
        'v' : ('test field', 'displacement', 'u'),
        'p' : ('unknown field', 'potential', 1, 1),
        'q' : ('test field', 'potential', 'p'),
        'pc' : ('unknown field', 'constant', 2, 1),
        'qc' : ('test field', 'constant', 'pc'),
    }

    materials = {
        'm' : ({
            'C': CE,
            'e' : e,
            'kappa' : kappa,
            'rho': rho,
            'penalty': 1,
            'iR' : 1.0 / (15e6 * dims[0] * dims[1]), # 1 / (R * top_area).
        },),
    }

    integrals = {
        'i' : 2 * order,
    }

    def get_ebcs(ts, coors, mode='u'):
        y = coors[:, 1]
        k = 2 * nm.pi / dims[1]
        shift = nm.pi / 3
        omega = cl * k
        time = ts.time
        if mode == 'u':
            val = (amplitude * nm.sin(time * omega) * nm.sin(k * y + shift))

        elif mode == 'du':
            val = (amplitude * omega * nm.cos(time * omega)
                   * nm.sin(k * y + shift))

        elif mode == 'ddu':
            val = (-amplitude * omega**2 * nm.sin(time * omega)
                   * nm.sin(k * y + shift))

        return val

    functions = {
        'get_u' : (lambda ts, coor, **kwargs: get_ebcs(ts, coor),),
        'get_du' : (lambda ts, coor, **kwargs: get_ebcs(ts, coor, mode='du'),),
        'get_ddu' : (lambda ts, coor, **kwargs: get_ebcs(ts, coor, mode='ddu'),),
    }

    ebcs = {
        'Seismic' : ('Bottom', {'u.2' : 'get_u', 'du.2' : 'get_du',
                                'ddu.2' : 'get_ddu'}),
        'Pot0' : ('Bottom', {'p.all' : 0.0}),
    }

    ics = {
        'ic' : ('Omega', {'u.all' : 0.0, 'du.all' : 0.0, 'p.0' : 0.0}),
    }

    equations = {
        '1' : """dw_dot.i.Omega(m.rho, v, ddu)
               + dw_lin_elastic.i.Omega(m.C, v, u)
               - dw_piezo_coupling.i.Omega(m.e, v, p)
               = 0""",
        '2' : """dw_piezo_coupling.i.Omega(m.e, u, q)
               + dw_diffusion.i.Omega(m.kappa, q, p)
               - de_surface_flux.i.Top(m.kappa, q, p)
               + de_surface_flux.i.Top(m.kappa, p, q)
               - de_surface_flux.i.Top(m.kappa, pc, q)
               + dw_dot.i.Top(m.penalty, q, p)
               - dw_dot.i.Top(m.penalty, q, pc)
               = 0""",
        '3' : """de_surface_flux.i.Top(m.kappa, qc, dp/dt)
               + 0.5 * dw_dot.i.Top(m.iR, qc, pc)
               + 0.5 * dw_dot.i.Top(m.iR, qc, pc[-1])
               = 0""",
    }

    solvers = {
        'lsd' : ('ls.auto_direct', {
            # Reuse the factorized linear system from the first time step.
            'use_presolve' : True,
            # Speed up the above by omitting the matrix digest check used
            # normally for verification that the current matrix corresponds to
            # the factorized matrix stored in the solver instance. Use with
            # care!
            'use_mtx_digest' : False,
            # Increase when getting MUMPS error -9.
            'memory_relaxation' : 20,
        }),
        'newton' : ('nls.newton', {
            'i_max'      : 1,
            'eps_a'      : 1e-6,
            'eps_r'      : 1e-6,
            'ls_on'      : 1e100,
        }),
        'tsn' : ('ts.newmark', {
            't0' : 0.0,
            't1' : t1,
            'dt' : dt,
            'n_step' : None,

            'is_linear' : True,
            # Without this the adaptive time-stepping cannot work.
            'has_time_derivatives' : adaptive,

            'beta' : 0.25,
            'gamma' : 0.5,

            'verbose' : 1,
        }),
        'tscedl' : ('tsc.ed_linear', {
            'eps_r' : (1e-4, 1e-2),
            'eps_a' : (1e-8, 1e-3),
            'fmin' : 0.3,
            'fmax' : 2.5,
            'fsafety' : 0.85,
            'red_factor' : 0.9,
            'inc_wait' : 10,
            'min_inc_factor' : 1.5,
        }),
    }

    options = {
        'ts' : tss_name,
        'tsc' : tsc_name if adaptive else None,
        'nls' : 'newton',
        'ls' : ls_name,

        'save_times' : save_times,

        'active_only' : active_only,
        'auto_transform_equations' : True,

        'output_format' : 'h5',
        'output_dir' : output_dir,
        'post_process_hook' : _post_process,
        'post_process_hook_final' : _plot_voltage,
    }

    return locals()



import sfepy.base.testing as tst

def test_tensors():
    import numpy as nm
    from sfepy.linalg import dot_sequences, insert_strided_axis

    ok = True

    a = nm.arange(1, 10).reshape(3, 3)
    b = nm.arange(9, 0, -1).reshape(3, 3)

    dab = nm.dot(a, b)
    dabt = nm.dot(a, b.T)
    datb = nm.dot(a.T, b)
    datbt = nm.dot(a.T, b.T)

    sa = insert_strided_axis(a, 0, 10)
    sb = insert_strided_axis(b, 0, 10)

    dsab = dot_sequences(sa, sb, mode='AB')
    _ok = nm.allclose(dab[None, ...], dsab, rtol=0.0, atol=1e-14)
    tst.report('dot_sequences AB: %s' % _ok)
    ok = ok and _ok

    dsabt = dot_sequences(sa, sb, mode='ABT')
    _ok = nm.allclose(dabt[None, ...], dsabt, rtol=0.0, atol=1e-14)
    tst.report('dot_sequences ABT: %s' % _ok)
    ok = ok and _ok

    dsatb = dot_sequences(sa, sb, mode='ATB')
    _ok = nm.allclose(datb[None, ...], dsatb, rtol=0.0, atol=1e-14)
    tst.report('dot_sequences ATB: %s' % _ok)
    ok = ok and _ok

    dsatbt = dot_sequences(sa, sb, mode='ATBT')
    _ok = nm.allclose(datbt[None, ...], dsatbt, rtol=0.0, atol=1e-14)
    tst.report('dot_sequences ATBT: %s' % _ok)
    ok = ok and _ok

    assert ok

def test_unique_rows():
    import numpy as nm
    from sfepy.linalg import unique_rows

    a = nm.arange(1, 10).reshape(3, 3)

    b = nm.r_[a, a]
    c = unique_rows(b)

    ok = (a == c).all()

    assert ok

def test_assemble1d():
    import numpy as nm
    from sfepy.linalg import assemble1d

    a = nm.arange(5)
    b = nm.arange(2)

    assemble1d(b, [1, 1, 1, 1, 0, 0], a[[0, 2, 3, 4, 1, 1]])

    ok = (b == [2, 10]).all()

    assert ok

def test_geometry():
    import numpy as nm
    from sfepy.linalg import get_face_areas

    a1 = get_face_areas([[0, 1, 2, 3]],
                        [[0, 0], [1, 0], [1, 1], [0, 1]])

    a2 = get_face_areas([[0, 1, 2, 3]],
                        [[0, 0, 2], [1, 0, 2], [1, 1, 2], [0, 1, 2]])
    ok = nm.allclose([a1, a2], [1, 1], rtol=0, atol=1e-15)

    assert ok

def test_get_blocks_stats():
    import numpy as nm
    from sfepy.linalg.utils import get_blocks_stats

    A = nm.eye(3)
    B = nm.full((3,2), 2)
    C = nm.full((1,3), 3)
    D = nm.full((1,2), 4)

    sr = [slice(0, 3), slice(3, 5)]
    sc = [slice(0, 3), slice(3, 4)]

    M = nm.block([[A, B], [C, D]])
    stats = get_blocks_stats(M, sr, sc)

    assert stats['shape'].tolist() == [[(3, 3), (3, 1)], [(1, 3), (1, 1)]]
    assert (stats['min'] == nm.array([[0., 2.], [3., 4.]])).all()
    assert (stats['max'] == nm.array([[1., 2.], [3., 4.]])).all()



#!/usr/bin/env python
"""
Convert a mesh file from one SfePy-supported format to another.
"""
from __future__ import absolute_import
import sys
import os.path as op
from ast import literal_eval
from six.moves import range
sys.path.append('.')

from argparse import ArgumentParser, RawDescriptionHelpFormatter
from sfepy.base.base import nm, output
from sfepy.base.ioutils import remove_files
from sfepy.discrete.fem import Mesh, FEDomain
from sfepy.discrete.fem.meshio import output_mesh_formats
from sfepy.discrete.fem.mesh import fix_double_nodes
import sfepy.mesh.mesh_tools as mt
from sfepy.mesh.mesh_generators import gen_tiled_mesh

helps = {
    'scale' : 'scale factor (float or comma-separated list for each axis)'
    ' [default: %(default)s]',
    'center' : 'center of the output mesh (0 for origin or'
    ' comma-separated list for each axis) applied after scaling'
    ' [default: %(default)s]',
    'refine' : 'uniform refinement level [default: %(default)s]',
    'format' : 'output mesh format (overrides filename_out extension)',
    'list' : 'list supported readable/writable output mesh formats',
    'merge' : 'remove duplicate vertices',
    'tri-tetra' : 'convert elements: quad->tri, hexa->tetra',
    '2d' : 'force a 2D mesh by removing the z coordinates - assumes a 3D mesh'
    ' in the xy plane',
    '3d' : 'force a 3D mesh by adding zero (y,) z coordinates',
    'cell-dim' : 'write only cells of a given dimension,'
    ' use a comma-separated list for several values',
    'save-per-mat': 'extract cells by material id and save them into'
    ' separate mesh files with a name based on filename_out and the id'
    ' numbers (preserves original mesh vertices)',
    'tile' : """
      scale a periodic input mesh (a rectangle or box) by a scale factor
      (--scale with a single integer value) and generate a new mesh by
      repeating the scaled original mesh in a regular grid (scale x scale [x
      scale]) if --tile=scale, or in a grid nx x ny[ x nz] for
      --tile=<nx,ny[,nz]>, producing again a periodic rectangle or box mesh.
      Use the --eps option to set the coordinate precision for periodicity
      checks.
    """,
    'extract_edges' : """
      extract outline edges of a given mesh and save it into a meshio-supported
      file. The outline edge is an edge for which norm(nvec1 - nvec2) < eps,
      where nvec1 and nvec2 are the normal vectors of the incident facets and
      eps is given by the --eps option.
    """,
    'eps' : """
      tolerance parameter [default: %(default)s]
    """,
    'extract_surface' : 'extract surface facets of a mesh',
    'print_surface' : """
       extract surface facets of a mesh and print it to a given file (use '-'
       for stdout) in form of a list where each row is [element, face,
       component]. A component corresponds to a contiguous surface region - for
       example, a cubical mesh with a spherical hole has two surface
       components. Two surface faces sharing a single node belong to one
       component.
    """,
    'remesh' : """when given, remesh the given mesh using tetgen.
      The options can be the following, separated by spaces, in this order: 1.
      "r" causes remeshing of the mesh volume - if not present the mesh surface
      is extracted and used for the volume mesh generation. 2.
      "q[<float>/<float>]" (required) - the two numbers after "q" are a maximum
      radius-edge ratio bound and a minimum dihedral angle bound. 3. "a<float>"
      (optional) - the number imposes a maximum volume constraint on all
      tetrahedra. 4. O[<0-9>/<0-7>] - the two numbers correspond to a mesh
      optimization level and a choice of optimizing operations. 5. "V"
      (optional) - if present, mesh statistics are printed. Consult the tetgen
      documentation for details. Examples: --remesh='rq2/0 a1e-8 O9/7 V'""",
}

def _parse_val_or_vec(option, name, parser):
    if option is not None:
        try:
            try:
                option = float(option)
            except ValueError:
                option = [float(ii) for ii in option.split(',')]
            option = nm.array(option, dtype=nm.float64, ndmin=1)
        except:
            output('bad %s! (%s)' % (name, option))
            parser.print_help()
            sys.exit(1)

    return option

def main():
    parser = ArgumentParser(description=__doc__,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('-s', '--scale', metavar='scale',
                        action='store', dest='scale',
                        default=None, help=helps['scale'])
    parser.add_argument('-c', '--center', metavar='center',
                        action='store', dest='center',
                        default=None, help=helps['center'])
    parser.add_argument('-r', '--refine', metavar='level',
                        action='store', type=int, dest='refine',
                        default=0, help=helps['refine'])
    parser.add_argument('-f', '--format', metavar='format',
                        action='store', type=str, dest='format',
                        default=None, help=helps['format'])
    parser.add_argument('-l', '--list', action='store_true',
                        dest='list', help=helps['list'])
    parser.add_argument('-m', '--merge', action='store_true',
                        dest='merge', help=helps['merge'])
    parser.add_argument('-t', '--tri-tetra', action='store_true',
                        dest='tri_tetra', help=helps['tri-tetra'])
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-2', '--2d', action='store_true',
                       dest='force_2d', help=helps['2d'])
    group.add_argument('-3', '--3d', action='store_true',
                       dest='force_3d', help=helps['3d'])
    parser.add_argument('-d', '--cell-dim', metavar='cell_dim',
                        action='store', dest='cell_dim',
                        default=None, help=helps['cell-dim'])
    parser.add_argument('--save-per-mat', action='store_true',
                        dest='save_per_mat', help=helps['save-per-mat'])
    parser.add_argument('--tile', metavar='nx,ny[,nz]', dest='tile',
                        default=None, help=helps['tile'])
    parser.add_argument('--extract-edges', action='store_true',
                        dest='extract_edges', help=helps['extract_edges'])
    parser.add_argument('--eps', action='store', type=float, dest='eps',
                        default=1e-12, help=helps['eps'])
    parser.add_argument('--extract-surface', action='store_true',
                        dest='extract_surface', help=helps['extract_surface'])
    parser.add_argument('--print-surface', action='store', metavar='filename',
                        dest='print_surface', default=None,
                        help=helps['print_surface'])
    parser.add_argument('--remesh', metavar='options',
                        action='store', dest='remesh',
                        default=None, help=helps['remesh'])
    parser.add_argument('filename_in')
    parser.add_argument('filename_out')
    options = parser.parse_args()

    if options.list:
        output('Supported readable mesh formats:')
        output('--------------------------------')
        output_mesh_formats('r')
        output('')
        output('Supported writable mesh formats:')
        output('--------------------------------')
        output_mesh_formats('w')
        sys.exit(0)

    scale = _parse_val_or_vec(options.scale, 'scale', parser)
    center = _parse_val_or_vec(options.center, 'center', parser)
    cell_dim = _parse_val_or_vec(options.cell_dim, 'cell_dim', parser)

    filename_in = options.filename_in
    filename_out = options.filename_out

    if options.remesh:
        import tempfile
        import shlex
        import subprocess

        dirname = tempfile.mkdtemp()

        is_surface = options.remesh.startswith('q')
        if is_surface:
            mesh = Mesh.from_file(filename_in)
            domain = FEDomain(mesh.name, mesh)
            region = domain.create_region('surf', 'vertices of surface',
                                          'facet')
            surf_mesh = Mesh.from_region(region, mesh,
                                         localize=True, is_surface=True)

            filename = op.join(dirname, 'surf.mesh')
            surf_mesh.write(filename, io='auto')

        else:
            import shutil

            shutil.copy(filename_in, dirname)
            filename = op.join(dirname, op.basename(filename_in))

        qopts = ''.join(options.remesh.split()) # Remove spaces.
        command = 'tetgen -BFENkACp%s %s' % (qopts, filename)
        args = shlex.split(command)
        subprocess.call(args)

        root, ext = op.splitext(filename)
        mesh = Mesh.from_file(root + '.1.vtk')

        remove_files(dirname)

    else:
        mesh = Mesh.from_file(filename_in)

    if cell_dim is not None:
        cell_dim = [int(ii) for ii in cell_dim]
        data = list(mesh._get_io_data(cell_dim_only=cell_dim))
        mesh = Mesh.from_data(mesh.name, *data)

    if options.force_2d:
        data = list(mesh._get_io_data(cell_dim_only=2))
        data[0] = data[0][:, :2]
        mesh = Mesh.from_data(mesh.name, *data)

    elif options.force_3d:
        data = list(mesh._get_io_data())
        data[0] = nm.pad(data[0], [(0, 0), (0, 3 - data[0].shape[1])])
        mesh = Mesh.from_data(mesh.name, *data)

    if scale is not None:
        if len(scale) == 1:
            tr = nm.eye(mesh.dim, dtype=nm.float64) * scale
        elif len(scale) == mesh.dim:
            tr = nm.diag(scale)
        else:
            raise ValueError('bad scale! (%s)' % scale)
        mesh.transform_coors(tr)

    if center is not None:
        cc = 0.5 * mesh.get_bounding_box().sum(0)
        shift = center - cc
        tr = nm.c_[nm.eye(mesh.dim, dtype=nm.float64), shift[:, None]]
        mesh.transform_coors(tr)

    if options.refine > 0:
        domain = FEDomain(mesh.name, mesh)
        output('initial mesh: %d nodes %d elements'
               % (domain.shape.n_nod, domain.shape.n_el))

        for ii in range(options.refine):
            output('refine %d...' % ii)
            domain = domain.refine()
            output('... %d nodes %d elements'
                   % (domain.shape.n_nod, domain.shape.n_el))

        mesh = domain.mesh

    if options.tri_tetra > 0:
        mesh = mt.triangulate(mesh, verbose=True)

    if options.merge:
        desc = mesh.descs[0]
        coor, ngroups, conns = fix_double_nodes(mesh.coors,
                                                mesh.cmesh.vertex_groups,
                                                mesh.get_conn(desc))
        mesh = Mesh.from_data(mesh.name + '_merged',
                              coor, ngroups,
                              [conns], [mesh.cmesh.cell_groups], [desc])

    if options.save_per_mat:
        desc = mesh.descs[0]
        conns, cgroups = mesh.get_conn(desc), mesh.cmesh.cell_groups
        coors, ngroups = mesh.coors, mesh.cmesh.vertex_groups
        mat_ids = nm.unique(cgroups)

        for mat_id in mat_ids:
            idxs = nm.where(cgroups == mat_id)[0]
            imesh = Mesh.from_data(mesh.name + '_matid_%d' % mat_id,
                                   coors, ngroups,
                                   [conns[idxs]], [cgroups[idxs]], [desc])

            fbase, fext = op.splitext(filename_out)
            ifilename_out = '%s_matid_%d%s' % (fbase, mat_id, fext)
            output('writing %s...' % ifilename_out)
            imesh.write(ifilename_out, file_format=options.format)
            output('...done')

    if options.tile:
        _scale = (scale[0] if isinstance(scale, nm.ndarray)
                  else scale if scale is not None else 1)
        repeat = (list(literal_eval(options.tile)) if options.tile != 'scale'
                  else None)
        output('tiling paramaters:')
        output('scale:', scale)
        output('repeat:', repeat)
        output('eps:', options.eps)

        mesh = gen_tiled_mesh(mesh, repeat, 1./_scale, options.eps)

    if options.extract_surface or options.print_surface:
        domain = FEDomain(mesh.name, mesh)

    if options.extract_surface:
        region = domain.create_region('surf', 'vertices of surface', 'facet')
        surf_mesh = Mesh.from_region(region, mesh,
                                     localize=True, is_surface=True)
        mesh = surf_mesh

    if options.print_surface:
        domain.fix_element_orientation()

        lst, surf_faces = mt.get_surface_faces(domain)

        gr_s = mt.surface_graph(surf_faces, domain.mesh.n_nod)

        n_comp, comps = mt.surface_components(gr_s, surf_faces)
        output('number of surface components:', n_comp)

        ccs, comps = comps, nm.zeros((0,1), nm.int32)
        for cc in ccs:
            comps = nm.concatenate((comps, cc[:,nm.newaxis]), 0)

        out = nm.concatenate((lst, comps), 1)
        if (options.print_surface == '-'):
            file_out = sys.stdout
        else:
            file_out = open(options.print_surface, 'w')
        for row in out:
            file_out.write('%d %d %d\n' % (row[0], row[1], row[2]))
        if (options.print_surface != '-'):
            file_out.close()

    if options.extract_edges:
        mesh_out = mt.extract_edges(mesh, eps=options.eps)
        mesh_out = mt.merge_lines(mesh_out)

        import meshio
        emesh = meshio.Mesh(mesh_out[0], [('line', mesh_out[2][0])],
                            cell_data={'mat_id' : mesh_out[3]})
        emesh.write(filename_out)

    else:
        output('writing %s...' % filename_out)
        mesh.write(filename_out, file_format=options.format, binary=False)
        output('...done')

if __name__ == '__main__':
    main()



r"""
A linear elastic beam loaded with a continuous force. The FE meshes consisting
of hexehedral, tetrahedral, and wedge elements are used in the simulation and
the results are compared.

The displacement at the beam end is compared to the reference
solution calculated on the homogeneous hexahedral mesh.

Running the simulation::

    sfepy-run sfepy/examples/linear_elasticity/wedge_mesh.py

Viewing the results::

    sfepy-view beam_h7.vtk beam_t42.vtk beam_w14.vtk -f u:s0:wu:e:p0 u:s1:wu:e:p0 u:s2:wu:e:p0 --camera-position="1.2,-0.6,0.1,0.4,0.1,-0.1,-0.2,0.1,1"
"""
import os.path as osp
import numpy as nm
from sfepy import data_dir
from sfepy.mechanics.matcoefs import stiffness_from_youngpoisson
from sfepy.base.base import Struct
from sfepy.discrete import Problem
from sfepy.discrete.fem import Mesh
from sfepy.discrete.fem.meshio import UserMeshIO
from scipy.spatial.transform import Rotation

global_dict = {}


def mesh_hook(mesh, mode):
    if mode == 'read' and 'mesh_hook_param' in global_dict:
        fname, angle = global_dict['mesh_hook_param']
        mesh = Mesh.from_file(fname)

        center = nm.average(mesh.cmesh.coors, axis=0)
        coors = mesh.cmesh.coors - center

        rot = Rotation.from_rotvec(nm.array([1., 0, 0]) * nm.deg2rad(angle))
        mesh.cmesh.coors[:] = rot.apply(coors) + center

        return mesh


def test_meshes(pb):
    out = []
    conf = pb.conf.copy()
    ok = True

    for mesh_group in meshes:
        displ = []
        for mesh in mesh_group:
            if isinstance(mesh, tuple):
                fname = osp.join(data_dir, 'meshes', '3d', mesh[0])
                angle = mesh[1]
                global_dict['mesh_hook_param'] = (fname, angle)
                conf.filename_mesh = UserMeshIO(mesh_hook)
            else:
                conf.filename_mesh = osp.join(data_dir, 'meshes', '3d', mesh)

            pb = Problem.from_conf(conf)

            yield pb, out

            displ.append(out[-1][1]['u'].data[-1].reshape((-1, 3)))

            yield None

        err = nm.array([d[-1, 2] - displ[0][-1, 2] for d in displ[1:]]) < 0.01
        ok = ok and err.all()

    print(f'wedge elements test: {["failed!", "passed"][int(ok)]}')


def get_force(ts, coors, mode=None, **kwargs):
    if mode == 'qp':
        force = 1e3

        val = nm.zeros_like(coors)[..., None]
        val[:, 2, 0] = -coors[:, 0] / 0.7 * force

        return {'val': val}


def post_proces(out, pb, state, extend=False):
    S = pb.evaluate('ev_cauchy_stress.i.Omega(solid.D, u)', mode='el_avg')
    out['stress'] = Struct(name='out', data=S, mode='cell')

    return out


meshes = [
    ['beam_h7.mesh', ('beam_w14.vtk', 90), ('beam_w14.vtk', 270)],
    ['beam_t42.mesh', 'beam_w14.vtk', ('beam_w14.vtk', 180)],
]

def define():
    filename_mesh = osp.join(data_dir, 'meshes', '3d', meshes[0][0])

    options = {
        'post_process_hook': 'post_proces',
        'parametric_hook': 'test_meshes',
        'output_dir': 'output',
    }

    regions = {
        'Omega': 'all',
        'Left': ('vertices in (x < 0.01)', 'facet'),
        'Right': ('vertices in (x > 0.69)', 'facet'),
        'Top': ('vertices in (z > 0.09)', 'facet'),
        'Bottom': ('vertices in (z < 0.01)', 'facet'),
        'Edge': ('r.Bottom *v r.Right', 'vertex'),
    }

    functions = {
        'get_force' : (get_force,),
    }

    materials = {
        'solid': ({'D': stiffness_from_youngpoisson(dim=3, young=1e6,
                                                    poisson=0.3)},),
        'force': 'get_force',
    }

    fields = {
        'displacement': ('real', 'vector', 'Omega', 1),
    }

    integrals = {
        'i': 2,
    }

    variables = {
        'u': ('unknown field', 'displacement', 0),
        'v': ('test field', 'displacement', 'u'),
    }

    ebcs = {
        'Fixed': ('Left', {'u.all' : 0.0}),
    }

    equations = {
        'balance_of_forces':
        """dw_lin_elastic.i.Omega(solid.D, v, u)
        = dw_surface_ltr.i.Top(force.val, v)"""
    }

    solvers = {
        'ls': ('ls.auto_direct', {}),
        'newton': ('nls.newton', {
            'i_max': 1,
            'eps_a': 1e-6,
        }),
    }

    return locals()



"""
Test continuity along a boundary with hanging nodes due to a mesh refinement.
"""
import os.path as op

import numpy as nm
import pytest

from sfepy.base.base import assert_, Struct
import sfepy.base.testing as tst

def eval_fun(ts, coors, mode, **kwargs):
    val = nm.sin(nm.sum(coors**2, axis=1) * nm.pi)

    val = val.reshape((coors.shape[0], 1, 1))
    return val

def _gen_lines_2_4(bbox, eps):
    from sfepy.discrete.probes import LineProbe

    LineProbe.__base__.cache = Struct(name='probe_shared_evaluate_cache')

    n_point = 100

    line0 = LineProbe([bbox[0, 0], -eps], [bbox[1, 0], -eps], n_point)
    line1 = LineProbe([bbox[0, 0], eps], [bbox[1, 0], eps], n_point)

    yield line0, line1

    line0 = LineProbe([-eps, bbox[0, 1]], [-eps, bbox[1, 1]], n_point)
    line1 = LineProbe([eps, bbox[0, 1]], [eps, bbox[1, 1]], n_point)

    yield line0, line1

def _gen_grid_3_8(bbox, eps):
    from sfepy.discrete.probes import PointsProbe

    PointsProbe.__base__.cache = Struct(name='probe_shared_evaluate_cache')

    n_point = 100

    px = nm.linspace(bbox[0, 0], bbox[1, 0], n_point)
    py = nm.linspace(bbox[0, 1], bbox[1, 1], n_point)
    pz = nm.linspace(bbox[0, 2], bbox[1, 2], n_point)

    # Test face substitutions.
    pps = nm.meshgrid(px, py, -eps)
    points0 = nm.array([ii.ravel() for ii in pps]).T
    pps = nm.meshgrid(px, py, eps)
    points1 = nm.array([ii.ravel() for ii in pps]).T

    grid0 = PointsProbe(points0)
    grid1 = PointsProbe(points1)

    yield grid0, grid1

    pps = nm.meshgrid(px, -eps, pz)
    points0 = nm.array([ii.ravel() for ii in pps]).T
    pps = nm.meshgrid(px, eps, pz)
    points1 = nm.array([ii.ravel() for ii in pps]).T

    grid0 = PointsProbe(points0)
    grid1 = PointsProbe(points1)

    yield grid0, grid1

    pps = nm.meshgrid(-eps, py, pz)
    points0 = nm.array([ii.ravel() for ii in pps]).T
    pps = nm.meshgrid(eps, py, pz)
    points1 = nm.array([ii.ravel() for ii in pps]).T

    grid0 = PointsProbe(points0)
    grid1 = PointsProbe(points1)

    yield grid0, grid1

    # Test edge substitutions.
    pps = nm.meshgrid(px, -eps, -eps)
    points0 = nm.array([ii.ravel() for ii in pps]).T
    pps = nm.meshgrid(px, eps, eps)
    points1 = nm.array([ii.ravel() for ii in pps]).T

    grid0 = PointsProbe(points0)
    grid1 = PointsProbe(points1)

    yield grid0, grid1

    pps = nm.meshgrid(-eps, py, -eps)
    points0 = nm.array([ii.ravel() for ii in pps]).T
    pps = nm.meshgrid(eps, py, eps)
    points1 = nm.array([ii.ravel() for ii in pps]).T

    grid0 = PointsProbe(points0)
    grid1 = PointsProbe(points1)

    yield grid0, grid1

    pps = nm.meshgrid(-eps, -eps, pz)
    points0 = nm.array([ii.ravel() for ii in pps]).T
    pps = nm.meshgrid(eps, eps, pz)
    points1 = nm.array([ii.ravel() for ii in pps]).T

    grid0 = PointsProbe(points0)
    grid1 = PointsProbe(points1)

    yield grid0, grid1

def _build_filenames(output_dir, key, order, ip):
    return [op.join(output_dir,
                   'test_refine_hanging_lin_%s_%d_%d.vtk' % (key, order, ip)),
            op.join(output_dir,
                    'test_refine_hanging_%s_%d_%d.vtk' % (key, order, ip))]

@pytest.fixture(scope='module')
def gels():
    from sfepy.discrete.fem.geometry_element import GeometryElement

    gels = {}
    for key in ['2_4', '3_8']:
        gel = GeometryElement(key)
        gel.create_surface_facet()
        gels[key] = gel

    return gels

@pytest.mark.slow
def test_continuity(gels, output_dir):
    from sfepy.base.base import Struct
    from sfepy.mesh.mesh_generators import gen_block_mesh
    from sfepy.discrete import FieldVariable
    from sfepy.discrete.fem import FEDomain, Field
    from sfepy.discrete.projections import make_l2_projection_data
    import sfepy.discrete.fem.refine_hanging as rh

    dims = [1.5, 2.0, 1.3]
    shape = [3, 3, 3]
    centre = [0.0, 0.0, 0.0]
    probe_gens = {'2_4' : _gen_lines_2_4, '3_8' : _gen_grid_3_8}

    ok = True
    for key, gel in gels.items():
        probe_gen = probe_gens[key]

        perms = gel.get_conn_permutations()

        dim = gel.dim
        for io, order in enumerate(range(1, 4)):
            mesh00 = gen_block_mesh(dims[:dim], shape[:dim], centre[:dim],
                                    name='block')

            for ip, perm in enumerate(perms):
                tst.report('geometry: %s, order: %d, permutation: %d: %s'
                           % (key, order, ip, perm))
                mesh0 = mesh00.copy()
                conn = mesh0.cmesh.get_conn(dim, 0).indices
                conn = conn.reshape((mesh0.n_el, -1))
                conn[-1, :] = conn[-1, perm]

                domain0 = FEDomain('d', mesh0)

                refine = nm.zeros(mesh0.n_el, dtype=nm.uint8)
                refine[:-1] = 1

                subs = None
                domain, subs = rh.refine(domain0, refine, subs=subs)

                omega = domain.create_region('Omega', 'all')
                field = Field.from_args('fu', nm.float64, 1, omega,
                                        approx_order=order)

                field.substitute_dofs(subs)

                uvar = FieldVariable('u', 'parameter', field,
                                     primary_var_name='(set-to-None)')

                field.restore_dofs(store=True)
                field.substitute_dofs(subs=None, restore=True)

                make_l2_projection_data(uvar, eval_fun)

                field.restore_dofs()

                bbox = domain.get_mesh_bounding_box()
                eps = 1e-7

                save = False
                for ii, (probe0, probe1) in enumerate(probe_gen(bbox, eps)):
                    probe0.set_options(close_limit=0.0)
                    probe1.set_options(close_limit=0.0)

                    pars0, vals0 = probe0(uvar)
                    pars1, vals1 = probe1(uvar)

                    assert_(nm.allclose(pars0, pars1, atol=1e-14, rtol=0.0))

                    _ok = nm.allclose(vals0, vals1, atol=20.0 * eps,
                                      rtol=0.0)
                    if not _ok:
                        save = True
                        tst.report('probe %d failed! (max. error: %e)'
                                   % (ii, nm.abs(vals0 - vals1).max()))

                    ok = ok and _ok

                if (ip == 0) or save:
                    out = uvar.create_output()
                    filenames = _build_filenames(output_dir,
                                                 key, order, ip)

                    domain.mesh.write(filenames[0], out=out)

                    linearization = Struct(kind='adaptive',
                                           min_level=0,
                                           max_level=4,
                                           eps=1e-2)

                    out = uvar.create_output(linearization=linearization)
                    val = out['u']
                    mesh = val.get('mesh', domain.mesh)
                    mesh.write(filenames[1], out=out)

    assert_(ok)

def test_preserve_coarse_entities(output_dir):
    from sfepy.mesh.mesh_generators import gen_block_mesh
    from sfepy.discrete.fem import FEDomain
    import sfepy.discrete.fem.refine_hanging as rh

    dims = [1.5, 2.0]
    shape = [11, 11]
    centre = [0.0, 0.0]

    mesh0 = gen_block_mesh(dims, shape, centre, name='block')
    domain0 = FEDomain('d', mesh0)
    reg = domain0.create_region('surface', 'vertices of surface', 'facet',
                                add_to_regions=False)
    cmesh0 = mesh0.cmesh

    cmesh0.vertex_groups[5::11] = 2
    cmesh0.vertex_groups[reg.vertices] = 1
    cmesh0.cell_groups[0] = 1
    cmesh0.cell_groups[50:60] = 2
    mesh0.write(op.join(output_dir,
                        'test_refine_hanging_ids0.vtk'), io='auto')

    refine = nm.zeros(mesh0.n_el, dtype=nm.uint8)
    refine[0:10] = 1
    refine[5::10] = 1

    domain, _, sub_cells = rh.refine(domain0, refine, subs=None,
                                     ret_sub_cells=True)
    domain.mesh.write(op.join(output_dir,
                              'test_refine_hanging_ids1.vtk'), io='auto')
    cmesh1 = domain.mesh.cmesh

    ii = nm.where(refine == 0)[0]
    conn0 = mesh0.get_conn('2_4')
    v0 = conn0[ii]

    conn1 = domain.mesh.get_conn('2_4')
    v1 = conn1[ii]

    ok = (v0 == v1).all()
    tst.report('coarse cells positions preserved:', ok)

    cgs0 = cmesh0.cell_groups[ii]
    cgs1 = cmesh1.cell_groups[ii]

    _ok = (cgs0 == cgs1).all()
    tst.report('coarse cells cell groups preserved:', _ok)
    ok = ok and _ok

    vgs0 = cmesh0.vertex_groups[v0]
    vgs1 = cmesh1.vertex_groups[v1]

    _ok = (vgs0 == vgs1).all()
    tst.report('coarse cells vertex groups preserved:', _ok)
    ok = ok and _ok

    ii = nm.where(refine == 1)[0]
    cgs0 = cmesh0.cell_groups[ii]
    cgs1 = cmesh1.cell_groups[sub_cells[:, 1:]]

    _ok = (cgs0[:, None] == cgs1).all()
    tst.report('refined cells cell groups preserved:', _ok)
    ok = ok and _ok

    assert_(ok)



# -*- coding: utf-8 -*-
r"""
Linear elasticity with multi node linear combination constraints.

Find :math:`\ul{u}` such that:

.. math::
    \int_{\Omega} D_{ijkl}\ e_{ij}(\ul{v}) e_{kl}(\ul{u})
    = 0
    \;, \quad \forall \ul{v} \;,

where

.. math::
    D_{ijkl} = \mu (\delta_{ik} \delta_{jl}+\delta_{il} \delta_{jk}) +
    \lambda \ \delta_{ij} \delta_{kl}
    \;.

The hanging nodes in ``'Dependent_nodes'`` region are bounded
to the nodes in ``'Independent_nodes'`` region using the ``lcbcs`
(``multi_node_combination``) conditions

View the results using::

  sfepy-view hanging_nodes.vtk -f u:wu:e -2
"""
import numpy as nm
from sfepy import data_dir
from sfepy.base.base import output
from sfepy.mechanics.matcoefs import stiffness_from_lame


def node_map_fun(coors0, coors1):
    """
    Map hanging nodes (coors0) and coincident edges (edge nodes in coors1)
    on a line.
    """

    ldir = coors1[-1] - coors1[0]
    ldir = ldir / nm.linalg.norm(ldir)
    if ldir[0] > 1e-9:
        par0 = coors0[:, 0] / ldir[0]
        par1 = coors1[:, 0] / ldir[0]
    else:
        par0 = coors0[:, 1] / ldir[1]
        par1 = coors1[:, 1] / ldir[1]

    idxs1 = nm.argsort(par1)
    nmap = []
    for par in par0:
        edge = nm.where(nm.logical_and(par1[:-1] < par, par < par1[1:]))[0][0]
        de = par1[edge + 1] - par1[edge]
        nmap.append((idxs1[edge], idxs1[edge + 1],
                     1 - (par - par1[edge]) / de,
                     1 - (par1[edge + 1] - par) / de))

    return (nm.arange(par0.shape[0]),
            nm.array([k[:2] for k in nmap]),
            nm.array([k[2:] for k in nmap]))


def check_lcs(out, pb, state, extend=None):
    displ = state.get_state_parts()['u'].reshape((-1, 2))
    coors = pb.domain.mesh.coors

    hnodes_map = {12: [2, 6], 14: [2, 6], 16: [6, 10], 18: [6, 10]}

    ok = True
    for nd, ends in hnodes_map.items():
        dy = coors[ends[1], 1] - coors[ends[0], 1]
        t = (coors[nd, 1] - coors[ends[0], 1]) / dy
        u_exp = displ[ends[0]] * (1 - t) + displ[ends[1]] * t
        ok = ok and (nm.linalg.norm(displ[nd] - u_exp) < 1e-9)

    result = {True: 'passed', False: 'failed'}[ok]
    output(f'multi node linear condition test {result}!')

    # tests_lcbcs hook:
    nls_status = pb.get_solver().status.nls_status
    if hasattr(nls_status, 'conditions'):
        nls_status.conditions.append(int(not ok))

    return out


filename_mesh = data_dir + '/meshes/2d/special/hanging_nodes.mesh'

options = {
    'post_process_hook': check_lcs,
}

regions = {
    'Omega': 'all',
    'Omega1': 'cells of group 1',
    'Omega2': 'cells of group 2',
    'Inside_nodes': ('vertices in (x > 0.199) & (x < 0.201)', 'vertex'),
    'Independent_nodes': ('r.Inside_nodes *v r.Omega1', 'vertex'),
    'Dependent_nodes': ('r.Inside_nodes -v r.Independent_nodes', 'vertex'),  
    'Left': ('vertices in (x < 0.001)', 'facet'),
    'Right': ('vertices in (x > 0.299)', 'facet'),
}

materials = {
    'solid': ({'D': stiffness_from_lame(dim=2, lam=1e1, mu=1e0)},),
}

fields = {
    'displacement': ('real', 'vector', 'Omega', 1),
}

integrals = {
    'i': 1,
}

variables = {
    'u': ('unknown field', 'displacement'),
    'v': ('test field', 'displacement', 'u'),
}

ebcs = {
    'Fixed': ('Left', {'u.all': 0.0}),
    'Displaced': ('Right', {'u.0': 0.05, 'u.1': -0.05}),
}

functions = {
    'node_map_fun': (node_map_fun,),
}

lcbcs = {
    'hanging': (['Dependent_nodes', 'Independent_nodes'], {'u.all': 'u.all'},
                'node_map_fun', 'multi_node_combination', None),
}

equations = {
    'balance_of_forces' :
    """dw_lin_elastic.i.Omega(solid.D, v, u) = 0""",
}

solvers = {
    'ls': ('ls.auto_direct', {}),
    'newton': ('nls.newton', {
        'i_max': 1,
        'eps_a': 1e-10,
    }),
}
